@@ -15,11 +15,12 @@ def stream_sqlite(sqlite_chunks, max_buffer_size):
     unsigned_long = Struct('>L')
     double = Struct('>d')
     table_leaf_header = Struct('>HHHB')
+    index_leaf_header = Struct('>HHHB')
     table_interior_header = Struct('>HHHBL')
     index_interior_header = Struct('>HHHBL')
     freelist_trunk_header = Struct('>LL')
 
-    master_row_constructor = namedtuple('MasterRow', ('type', 'name', 'tbl_name', 'rootpage', 'sql'))
+    master_row_constructor = namedtuple('MasterRow', ('rowid', 'type', 'name', 'tbl_name', 'rootpage', 'sql'))
     column_constructor = namedtuple('Column', ('cid', 'name', 'type', 'notnull', 'dflt_value', 'pk'))
 
     def get_byte_reader(iterable):
@@ -156,62 +157,58 @@ def stream_sqlite(sqlite_chunks, max_buffer_size):
             nonlocal num_bytes_buffered
             num_bytes_buffered -= num_bytes
 
+        def process_initial_payload(initial_payload_size,
+                                    full_payload_size, full_payload_processor,
+                                    cell_num_reader, cell_varint_reader):
+            if initial_payload_size == full_payload_size:
+                yield from full_payload_processor(cell_num_reader, cell_varint_reader)
+
+            else:
+                initial_payload = cell_num_reader(initial_payload_size)
+                overflow_page, = unsigned_long.unpack(cell_num_reader(4))
+                payload_chunks = deque()
+                note_increase_buffered(len(initial_payload))
+                payload_chunks.append(initial_payload)
+                payload_remainder = full_payload_size - initial_payload_size
+
+                yield from process_if_buffered_or_remember(partial(
+                    process_overflow_page,
+                    full_payload_processor, payload_chunks, payload_remainder
+                ), overflow_page)
+
+        def process_overflow_page(full_payload_processor, payload_chunks, payload_remainder, page_bytes, page_reader):
+            next_overflow_page, = unsigned_long.unpack(page_reader(4))
+            num_this_page = min(payload_remainder, len(page_bytes) - 4)
+            payload_remainder -= num_this_page
+            note_increase_buffered(num_this_page)
+            payload_chunks.append(page_reader(num_this_page))
+
+            if not next_overflow_page:
+                payload = b''.join(payload_chunks)
+                note_decrease_buffered(len(payload))
+                yield from full_payload_processor(*get_chunk_readers(payload))
+
+            else:
+                yield from process_if_buffered_or_remember(partial(
+                    process_overflow_page,
+                    full_payload_processor, payload_chunks, payload_remainder
+                ), next_overflow_page)
+
         def process_table_page(table_name, table_info, row_constructor, page_bytes, page_reader):
 
-            def process_initial_payload(full_payload_processor, cell_num_reader, cell_varint_reader):
+            def get_table_initial_payload_size(p):
+                u = len(page_bytes)
+                x = u - 35
+                m = 32 * (u - 12) // 255 - 23
+                k = m + (p - m) % (u - 4)
 
-                def get_initial_payload_size(p):
-                    u = len(page_bytes)
-                    x = u - 35
-                    m = 32 * (u - 12) // 255 - 23
-                    k = m + (p - m) % (u - 4)
+                return (
+                    p if p <= x else \
+                    k if k <= x else \
+                    m
+                )
 
-                    return (
-                        p if p <= x else \
-                        k if k <= x else \
-                        m
-                    )
-
-                full_payload_size, _ = cell_varint_reader()
-                rowid, _ = cell_varint_reader()
-
-                initial_payload_size = get_initial_payload_size(full_payload_size)
-
-                if initial_payload_size == full_payload_size:
-                    yield from full_payload_processor(cell_num_reader, cell_varint_reader)
-
-                else:
-                    initial_payload = cell_num_reader(initial_payload_size)
-                    overflow_page, = unsigned_long.unpack(cell_num_reader(4))
-                    payload_chunks = deque()
-                    note_increase_buffered(len(initial_payload))
-                    payload_chunks.append(initial_payload)
-                    payload_remainder = full_payload_size - initial_payload_size
-
-                    yield from process_if_buffered_or_remember(partial(
-                        process_overflow_page,
-                        full_payload_processor, payload_chunks, payload_remainder
-                    ), overflow_page)
-
-            def process_overflow_page(full_payload_processor, payload_chunks, payload_remainder, page_bytes, page_reader):
-                next_overflow_page, = unsigned_long.unpack(page_reader(4))
-                num_this_page = min(payload_remainder, len(page_bytes) - 4)
-                payload_remainder -= num_this_page
-                note_increase_buffered(num_this_page)
-                payload_chunks.append(page_reader(num_this_page))
-
-                if not next_overflow_page:
-                    payload = b''.join(payload_chunks)
-                    note_decrease_buffered(len(payload))
-                    yield from full_payload_processor(*get_chunk_readers(payload))
-
-                else:
-                    yield from process_if_buffered_or_remember(partial(
-                        process_overflow_page,
-                        full_payload_processor, payload_chunks, payload_remainder
-                    ), next_overflow_page)
-
-            def read_table_row(cell_num_reader, cell_varint_reader):
+            def read_table_row(rowid, cell_num_reader, cell_varint_reader):
 
                 def serial_types(header_remaining, cell_varint_reader):
                     while header_remaining:
@@ -222,7 +219,7 @@ def stream_sqlite(sqlite_chunks, max_buffer_size):
                 header_remaining, header_varint_size = cell_varint_reader()
                 header_remaining -= header_varint_size
 
-                return row_constructor(*tuple(
+                return row_constructor(rowid, *tuple(
                     parser(cell_num_reader(length))
                     for serial_type in tuple(serial_types(header_remaining, cell_varint_reader))
                     for (length, parser) in (
@@ -248,11 +245,11 @@ def stream_sqlite(sqlite_chunks, max_buffer_size):
                     cur.execute(master_row.sql)
                     cur.execute("PRAGMA table_info('" + master_row.name.replace("'","''") + "');")
                     columns = cur.fetchall()
-                    row_constructor = namedtuple('Row', (column[1] for column in columns))
+                    row_constructor = namedtuple('Row', ('rowid',) + tuple(column[1] for column in columns))
                     return tuple(column_constructor(*column) for column in columns), row_constructor
 
-                def process_master_leaf_row(cell_num_reader, cell_varint_reader):
-                    master_row = read_table_row(cell_num_reader, cell_varint_reader)
+                def process_master_leaf_row(rowid, cell_num_reader, cell_varint_reader):
+                    master_row = read_table_row(rowid, cell_num_reader, cell_varint_reader)
                     yield from (
                         process_if_buffered_or_remember(partial(process_table_page, master_row.name, *table_info_and_row_constructor(cur, master_row)), master_row.rootpage) if master_row.type == 'table' else \
                         process_if_buffered_or_remember(process_index_page, master_row.rootpage) if master_row.type == 'index' else \
@@ -268,15 +265,18 @@ def stream_sqlite(sqlite_chunks, max_buffer_size):
 
                     for pointer, in pointers:
                         cell_num_reader, cell_varint_reader = get_chunk_readers(page_bytes, pointer)
+                        full_payload_size, _ = cell_varint_reader()
+                        rowid, _ = cell_varint_reader()
+                        initial_payload_size = get_table_initial_payload_size(full_payload_size)
                         yield from process_initial_payload(
-                            process_master_leaf_row,
+                            initial_payload_size, full_payload_size, partial(process_master_leaf_row, rowid),
                             cell_num_reader, cell_varint_reader,
                         )
 
             def process_table_leaf_non_master():
 
-                def process_non_master_leaf_row(cell_num_reader, cell_varint_reader):
-                    yield table_name, table_info, read_table_row(cell_num_reader, cell_varint_reader)
+                def process_non_master_leaf_row(rowid, cell_num_reader, cell_varint_reader):
+                    yield table_name, table_info, read_table_row(rowid, cell_num_reader, cell_varint_reader)
 
                 _, num_cells, _, _ = table_leaf_header.unpack(page_reader(7))
 
@@ -284,8 +284,12 @@ def stream_sqlite(sqlite_chunks, max_buffer_size):
 
                 for pointer, in pointers:
                     cell_num_reader, cell_varint_reader = get_chunk_readers(page_bytes, pointer)
+                    full_payload_size, _ = cell_varint_reader()
+                    rowid, _ = cell_varint_reader()
+                    initial_payload_size = get_table_initial_payload_size(full_payload_size)
+
                     yield from process_initial_payload(
-                        process_non_master_leaf_row,
+                        initial_payload_size, full_payload_size, partial(process_non_master_leaf_row, rowid),
                         cell_num_reader, cell_varint_reader,
                     )
 
@@ -313,10 +317,42 @@ def stream_sqlite(sqlite_chunks, max_buffer_size):
 
         def process_index_page(page_bytes, page_reader):
 
+            def get_index_initial_payload_size(p):
+                u = len(page_bytes)
+                x = (u - 12 ) * 64 // 255 -23
+                m = 32 * (u - 12) // 255 - 23
+                k = m + (p - m) % (u - 4)
+
+                return (
+                    p if p <= x else \
+                    k if k <= x else \
+                    m
+                )
+
             def process_index_leaf():
-                yield from ()
+
+                def process_index_leaf_row(cell_num_reader, cell_varint_reader):
+                    yield from ()
+
+                _, num_cells, _, _ = index_leaf_header.unpack(page_reader(7))
+
+                pointers = unsigned_short.iter_unpack(page_reader(num_cells * 2))
+
+                for pointer, in pointers:
+                    cell_num_reader, cell_varint_reader = get_chunk_readers(page_bytes, pointer)
+                    full_payload_size, _ = cell_varint_reader()
+                    initial_payload_size = get_index_initial_payload_size(full_payload_size)
+
+                    yield from process_initial_payload(
+                        initial_payload_size, full_payload_size, process_index_leaf_row,
+                        cell_num_reader, cell_varint_reader,
+                    )
 
             def process_index_interior():
+
+                def process_index_interior_row(cell_num_reader, cell_varint_reader):
+                    yield from ()
+
                 _, num_cells, _, _, right_most_pointer = \
                     index_interior_header.unpack(page_reader(11))
 
@@ -327,6 +363,13 @@ def stream_sqlite(sqlite_chunks, max_buffer_size):
                     page_num, =  unsigned_long.unpack(cell_num_reader(4))
                     yield from process_if_buffered_or_remember(process_index_page, page_num)
 
+                    full_payload_size, _ = cell_varint_reader()
+                    initial_payload_size = get_index_initial_payload_size(full_payload_size)
+                    yield from process_initial_payload(
+                        initial_payload_size, full_payload_size, process_index_interior_row,
+                        cell_num_reader, cell_varint_reader,
+                    )
+
                 yield from process_if_buffered_or_remember(process_index_page, right_most_pointer)
 
             page_type = page_reader(1)
@@ -371,15 +414,12 @@ def stream_sqlite(sqlite_chunks, max_buffer_size):
             else:
                 yield from process_page(page_bytes, page_reader)
 
-        if len(page_buffer) != 0:
-            raise ValueError('Unidentified pages in buffer')
+        if num_bytes_buffered != 0:
+            raise ValueError('Bytes remain in cache')
 
         if len(page_processors) != 0:
             raise ValueError("Expected a page that wasn't processed")
 
-        if num_bytes_buffered != 0:
-            raise ValueError('Bytes remain in cache')
-
     get_bytes = get_byte_reader(sqlite_chunks)
     page_size, num_pages_expected, first_freelist_trunk_page, incremental_vacuum = parse_header(get_bytes(100))
 
